# -*- coding: utf-8 -*-
from appdirs import AppDirs

_dirs = AppDirs("packratt")
user_data_dir = _dirs.user_data_dir
user_cache_dir = _dirs.user_cache_dir
user_config_dir = _dirs.user_config_dir

del _dirs



import logging
from pathlib import Path
import shutil

from packratt.cache import validate_entry, get_cache, set_cache
from packratt.downloads import downloaders

log = logging.getLogger(__name__)


def get(key, destination, entry=None):
    """
    Parameters
    ----------
    key : str
        Retrieval key
    destination : str or Path
        Path in which to unarchive the data product
    entry : None or dict
        dictionary entry describing the data product, if not in the registry.
        Defaults to None in which case the entry must be in the registry.
    """

    cache = get_cache()

    if entry is None:
        try:
            entry = cache[key]
        except KeyError:
            raise ValueError("%s is not in the registry" % key)
    elif isinstance(entry, dict):
        validate_entry(entry)
        cache[key] = entry
        set_cache(cache)
        entry = cache[key]
    else:
        raise TypeError("entry must be None or dict")

    if not isinstance(destination, Path):
        destination = Path(destination)

    if destination.exists() and not destination.is_dir():
        # Don't overwrite existing files
        raise ValueError(
            "Target %s exists and is not a directory" % destination)
    else:
        # Create the destination directory
        destination.mkdir(parents=True, exist_ok=True)

    filename = entry['dir'] / Path(key).name

    if filename.exists():
        # TODO(sjperkins):
        # This is a massive assumption, what about partial downloads etc.
        sha256_hash = entry['hash']
    else:
        # Download to the destination
        sha256_hash = downloaders(entry['type'], key, entry)

        if not sha256_hash == entry['hash']:
            raise ValueError("sha256hash does not agree. %s vs %s"
                             % (sha256_hash, entry['hash']))

    shutil.unpack_archive(str(filename), destination)

    return sha256_hash



# -*- coding: utf-8 -*-
from hashlib import sha224
from pathlib import Path
from threading import Lock
import weakref

from jsonschema import validate, ValidationError
import yaml

from packratt.directories import user_cache_dir
from packratt.registry import load_registry, ENTRY_SCHEMA


def validate_entry(entry):
    try:
        return validate(entry, ENTRY_SCHEMA)
    except ValidationError as e:
        raise ValueError("%s is not a valid entry" % entry) from e


_cache_lock = Lock()
_cache_cache = weakref.WeakValueDictionary()


class CacheMetaClass(type):
    """
    https://en.wikipedia.org/wiki/Multiton_pattern

    """
    def __call__(cls, cache_dir):
        with _cache_lock:
            try:
                return _cache_cache[cache_dir]
            except KeyError:
                instance = type.__call__(cls, cache_dir)
                _cache_cache[cache_dir] = instance
                return instance


class Cache(metaclass=CacheMetaClass):
    def __init__(self, cache_dir: Path):
        self.cache_dir = cache_dir
        self.registry = load_registry()

    def __reduce__(self):
        return (Cache, (self.cache_dir,))

    @staticmethod
    def key_dir(key) -> Path:
        h = sha224(key.encode('ascii')).hexdigest()
        return Path(h[0:2], h[2:4], h[4:6], h[6:])

    def cache_key_dir(self, key) -> Path:
        """
        Parameters
        ----------
        key : str
            Cache entry key

        Returns:
        path : :class:`Path`
            Path to the cache entry
        """
        return Path(self.cache_dir) / self.key_dir(key)

    def __setitem__(self, key, value):
        """
        Inserts a Cache entry

        Parameters
        ----------
        key : str
            Cache entry key
        value : dict
            Cache entry
        """
        if not isinstance(value, dict):
            raise TypeError("value must be a dict")

        # Is this a valid cache entry dictionary?
        validate_entry(value)

        entry_dir = self.cache_key_dir(key)

        try:
            entry_dir.mkdir(parents=True, exist_ok=True)
        except FileExistsError as e:
            raise ValueError("Already exists") from e

        with open(entry_dir / "entry.yaml", "w") as f:
            f.write(yaml.safe_dump(value))

    def __getitem__(self, key):
        """
        Retrieves a cache entry

        Parameters
        ----------
        key : str
            Cache entry key

        Returns
        -------
        entry : :class:`CacheEntry`
            Cache entry
        """
        entry_dir = self.cache_key_dir(key)

        # Try obtain the entry from the registry first
        # It's the primary source of truth
        try:
            entry = self.registry[key]
        except KeyError:
            # Look for an entry file as the secondary source of truth
            entry_file = entry_dir / "entry.yaml"

            # Nothing, raise a KeyError
            if not entry_file.exists():
                raise KeyError(key)

            with open(entry_file, "r") as f:
                entry = yaml.safe_load(f)
        else:
            self.__setitem__(key, entry)

        entry['dir'] = entry_dir

        return entry


def cache_factory(cache_dir=None):
    if cache_dir is None:
        return Cache(cache_dir)

    return Cache(user_cache_dir)


app_cache = None


def get_cache():
    global app_cache

    if app_cache is None:
        app_cache = Cache(user_cache_dir)

    return app_cache


def set_cache(cache):
    global app_cache

    if not isinstance(cache, Cache):
        raise TypeError("cache must be of type Cache")

    app_cache = cache



# -*- coding: utf-8 -*-
import argparse
import logging
import sys

from packratt.dispatch import Dispatch

log = logging.getLogger(__name__)

logging.basicConfig(level=logging.DEBUG)

commands = Dispatch()


@commands.register("get")
def get(args):
    from packratt.interface import get
    return get(args.key, args.destination)


def create_parser():
    p = argparse.ArgumentParser()
    sp = p.add_subparsers(help='command', dest='command')

    # Get a data product
    get = sp.add_parser('get')
    get.add_argument("key")
    get.add_argument("destination", default=".", nargs="?")

    return p


def _run(args):
    parser = create_parser()
    args = parser.parse_args(args)

    if args.command:
        commands(args.command, args)
    else:
        parser.print_help()

    return 0


def run():
    return _run(sys.argv[1:])


if __name__ == "__main__":
    sys.exit(run())



# -*- coding: utf-8 -*-
import uuid

import pickle

from packratt.cache import Cache


def test_cache_multiton(tmp_path_factory):
    cache_dir = tmp_path_factory.mktemp("cache")
    cache_dir2 = tmp_path_factory.mktemp("cache2")
    # Test that we get the same object when creating with the same cache_dir
    assert Cache(cache_dir) is Cache(cache_dir)

    # Test that we get the same object when unpickling with the same cache_dir
    assert pickle.loads(pickle.dumps(Cache(cache_dir))) is Cache(cache_dir)

    # Sanity check
    assert Cache(cache_dir) is not Cache(cache_dir2)


def test_cache(test_cache):
    key = "/ms/wsrt/wst/ms.tar.gz"
    entry = {"url": "ftp://elwood.ru.ac.za/pub/sjperkins/wsrt.ms.tar.gz",
             "hash": uuid.uuid4().hex,
             "filename": "wsrt.ms.tar.gz"}

    test_cache[key] = entry
    cmp_entry = entry.copy()
    cmp_entry['dir'] = test_cache.cache_dir / test_cache.key_dir(key)

    assert cmp_entry == test_cache[key]



# -*- coding: utf-8 -*-

from pathlib import Path

import yaml
from jsonschema import validate

import packratt
from packratt.directories import user_config_dir

SCHEMA_PATH = Path(Path(packratt.__file__).parent,
                   "conf", "registry-schema.yaml")

# Open the schema file
with open(SCHEMA_PATH, "r") as f:
    SCHEMA = yaml.safe_load(f)

# Create registry and entry schemas
REGISTRY_SCHEMA = {**SCHEMA, "$ref": "#/definitions/registry"}
ENTRY_SCHEMA = {**SCHEMA, "entry": {"type": {"$ref": "/definitions/entry"}}}

del SCHEMA_PATH


USER_REGISTRY = Path(user_config_dir, "registry.yaml")


def load_registry(filename=None):
    path = Path(Path(packratt.__file__).parent, "conf", "registry.yaml")

    with open(path, "r") as f:
        registry = yaml.safe_load(f)

    if USER_REGISTRY.is_file():
        with open(USER_REGISTRY, "r") as f:
            registry.update(yaml.safe_load(f))

    validate(registry, schema=REGISTRY_SCHEMA)

    return registry



from contextlib import contextmanager
from ftplib import FTP
import hashlib
import logging
from pathlib import Path
import shutil
from urllib.parse import urlparse

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from packratt.dispatch import Dispatch

log = logging.getLogger(__name__)

downloaders = Dispatch()

# 32k chunks
CHUNK_SIZE = 2**15


@contextmanager
def retry_strategy():
    retry = Retry(
        total=5,
        status_forcelist=[413, 429, 500, 502, 503, 504],
        method_whitelist=["HEAD", "GET", "OPTIONS"]
    )
    adapter = HTTPAdapter(max_retries=retry)
    session = requests.Session()
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    try:
        yield session
    finally:
        session.close()


@contextmanager
def open_and_hash_file(filename):
    sha256hash = hashlib.sha256()
    size = 0

    if filename.is_file():
        # File exists, hash contents and determine existing size
        # This also moves the file pointer to the end of the file
        f = open(filename, "rb+")

        while True:
            chunk = f.read(CHUNK_SIZE)

            if not chunk:
                break

            sha256hash.update(chunk)
            size += len(chunk)
    else:
        # Open a new file for writing
        f = open(filename, "wb")

    try:
        yield size, sha256hash, f
    finally:
        f.close()


def requests_partial_download(key, entry, url, session,
                              response, params=None):
    filename = Path(key).name
    part_filename = entry['dir'] / '.'.join((filename, 'partial'))
    filename = entry['dir'] / filename

    if params is None:
        params = {}

    with open_and_hash_file(part_filename) as (size, sha256hash, f):
        if size > 0:
            log.info("Resuming download of %s at %d", filename, size)
            # Some of this file has already been downloaded
            # Request the rest of it
            total_size = int(response.headers['Content-Length'])
            response.close()
            headers = {'Range': 'bytes=%d-%d' % (size, total_size)}
            response = session.get(url, params=params,
                                   headers=headers, stream=True)

        if response.status_code != 416:
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as e:
                raise Exception(e)

            for chunk in response.iter_content(CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    sha256hash.update(chunk)
                    f.write(chunk)

    shutil.move(part_filename, filename)
    return sha256hash.hexdigest()


@downloaders.register("google")
def download_google_drive(key, entry):
    url = "https://drive.google.com/uc?export=download"
    params = {'id': entry['file_id']}

    with retry_strategy() as session:
        response = session.get(url, params=params, stream=True)

        try:
            # Look for a token indicating a large file
            # Re-request with confirmation token
            for k, v in response.cookies.items():
                if k.startswith('download_warning'):
                    params['confirm'] = v
                    response.close()
                    response = session.get(url, params=params, stream=True)
                    break

            return requests_partial_download(key, entry, url,
                                             session, response,
                                             params=params)
        finally:
            response.close()


def download_ftp(key, entry, url):
    filename = Path(key).name
    part_filename = entry['dir'] / '.'.join((filename, 'partial'))
    filename = entry['dir'] / filename

    with open_and_hash_file(part_filename) as (size, sha256hash, f):
        if size > 0:
            log.info("Resuming download of %s at %d", part_filename, size)

        ftp = FTP(url.hostname)

        def callback(data):
            f.write(data)
            sha256hash.update(data)

        try:
            ftp.login(url.username, url.password)
            ftp.retrbinary("RETR %s" % url.path, callback,
                           blocksize=CHUNK_SIZE, rest=size)

            shutil.move(part_filename, filename)
        finally:
            ftp.quit()

        return sha256hash.hexdigest()


@downloaders.register("url")
def download_url(key, entry):
    # requests doesn't handle ftp
    parsed_url = urlparse(entry['url'])

    if parsed_url.scheme == "ftp":
        return download_ftp(key, entry, parsed_url)

        # Use requests for (presumably) http cases
        with retry_strategy() as session:
            response = session.get(entry['url'], stream=True)

            try:
                return requests_partial_download(key, entry, entry['url'],
                                                 session, response)
            finally:
                response.close()



class Dispatch:
    def __init__(self):
        self.lookup = {}

    def register(self, key, fn=None):
        def wrapper(fn):
            self.lookup[key] = fn

            return fn

        return wrapper(fn) if fn is not None else wrapper

    def __call__(self, key, *args, **kwargs):
        try:
            fn = self.lookup[key]
        except KeyError:
            raise ValueError("%s not registered on the dispatcher" % key)
        else:
            return fn(*args, **kwargs)



from packratt.cache import get_cache, set_cache, Cache
from packratt.registry import load_registry

from unittest.mock import patch
import pytest
import yaml


@pytest.fixture(scope="session", autouse=True)
def global_test_cache(tmp_path_factory):
    cache_dir = tmp_path_factory.mktemp('cache')
    set_cache(Cache(cache_dir))


@pytest.fixture
def test_cache(tmp_path_factory):
    cache_dir = tmp_path_factory.mktemp('cache')

    old_cache = get_cache()

    try:
        new_cache = Cache(cache_dir)
        set_cache(new_cache)
        yield new_cache
    finally:
        set_cache(old_cache)


USER_REGISTRY_CONTENT = {
    '/test/ms/2020-06-04/google/test_user_registry_ms.tar.gz': {
        'type': 'google',
        'file_id': '1wjZoh7OAIVEjYuTmg9dLAFiLoyehnIcL',
        'hash': '4d548b22331fb3cd3256b1b4f37a41cf',
        'description': 'Small testing Measurement Set, stored on Google Drive',
    }
}


@pytest.fixture(scope="session")
def registry(tmp_path_factory):
    conf = tmp_path_factory.mktemp('conf') / "registry.yaml"

    with open(conf, "w") as f:
        f.write(yaml.safe_dump(USER_REGISTRY_CONTENT))

    with patch('packratt.registry.USER_REGISTRY', conf):
        return load_registry()



# -*- coding: utf-8 -*-

# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

readme = ''

setup(
    long_description=readme,
    name='packratt',
    version='0.1.2',
    python_requires='==3.*,>=3.6.0',
    author='Simon Perkins',
    author_email='simon.perkins@gmail.com',
    entry_points={"console_scripts": ["packratt = packratt.application:run"]},
    packages=['packratt', 'packratt.tests'],
    package_dir={"": "."},
    package_data={"packratt": ["conf/*.yaml"]},
    install_requires=[
        'appdirs==1.*,>=1.4.4', 'jsonschema==3.*,>=3.2.0',
        'pyyaml==5.*,>=5.3.1', 'requests==2.*,>=2.23.0'
    ],
    extras_require={
        "dev": ["pylint==2.*,>=2.5.3"],
        "testing": [
            "pytest[testing]==5.*,>=5.2.0,>=5.4.2",
            "pytest-flake8[testing]==1.*,>=1.0.6"
        ]
    },
)



from packratt.interface import get

from pathlib import Path
import pytest
import shutil


@pytest.mark.skip
@pytest.mark.parametrize(
    "google_key", ['1.5M_water.tar.gz'])
def test_get_entry(google_key, tmp_path_factory):
    entry = {
        "type": 'google',
        "file_id": '1D62EwpZOL7I5MBVh5sskj7w9e-YaCSzX',
        "hash": ('3d6ab84b5ce54e8ac5b4e783458caf2a'
                 'faaf5e8e3cca3ee082ae431498bd4b37'),
        "description": '1.5M water'}
    google_entry_dest = tmp_path_factory.mktemp("google")
    google_entry_sha256 = get(google_key, google_entry_dest, entry=entry)

    assert google_entry_sha256 == entry['hash']


@pytest.mark.parametrize(
    "elwood_key", ['/test/ms/2020-06-04/elwood/smallest_ms.tar.gz'])
@pytest.mark.parametrize(
    "google_key", ['/test/ms/2020-06-04/google/smallest_ms.tar.gz'])
def test_get(google_key, elwood_key, test_cache, tmp_path_factory):
    google_dest = tmp_path_factory.mktemp("google")
    elwood_dest = tmp_path_factory.mktemp("elwood")

    google_sha256 = get(google_key, google_dest)
    elwood_sha256 = get(elwood_key, elwood_dest)

    assert google_sha256 == elwood_sha256


@pytest.mark.parametrize(
    "partial_key", ['/test/ms/2020-06-04/elwood/smallest_ms_truncated.tar.gz'])
@pytest.mark.parametrize(
    "elwood_key", ['/test/ms/2020-06-04/elwood/smallest_ms.tar.gz'])
@pytest.mark.parametrize(
    "google_key", ['/test/ms/2020-06-04/google/smallest_ms.tar.gz'])
def test_get_partial(partial_key, google_key, elwood_key,
                     test_cache, registry, tmp_path_factory):
    sha256 = get(partial_key, tmp_path_factory.mktemp("ignore"))
    assert sha256 == registry[partial_key]['hash']

    partial_file = Path(test_cache.cache_key_dir(partial_key),
                        Path(partial_key).name)

    # Create the cache directory for the full file
    elwood_dir = test_cache.cache_key_dir(elwood_key)
    elwood_dir.mkdir(parents=True, exist_ok=True)
    # Create the partial file
    partial_name = elwood_dir / '.'.join((Path(elwood_key).name, 'partial'))
    partial_path = Path(partial_name)
    full_path = Path(elwood_dir / Path(elwood_key).name)
    shutil.copyfile(partial_file, partial_path)

    dest = tmp_path_factory.mktemp("dest")

    assert not full_path.exists()
    assert partial_path.exists()
    assert get(elwood_key, dest) == registry[elwood_key]['hash']
    assert full_path.exists()
    assert not partial_path.exists()

    # Create the cache directory for the full file
    google_dir = test_cache.cache_key_dir(google_key)
    google_dir.mkdir(parents=True, exist_ok=True)
    # Create the partial file
    partial_name = google_dir / '.'.join((Path(google_key).name, 'partial'))
    partial_path = Path(partial_name)
    full_path = Path(google_dir / Path(google_key).name)
    shutil.copyfile(partial_file, partial_name)

    dest = tmp_path_factory.mktemp("dest")

    assert not full_path.exists()
    assert partial_path.exists()
    assert get(google_key, dest) == registry[google_key]['hash']
    assert full_path.exists()
    assert not partial_path.exists()



from jsonschema import validate

from packratt.registry import REGISTRY_SCHEMA
from packratt.tests.conftest import USER_REGISTRY_CONTENT


def test_schema():
    validate(
        {
            "/ms/vla/vla012345.ms": {
                "type": "url",
                "url": "https://blah.org/vla012345.ms",
                "hash": "123456",
                "filename": "vla012345.ms",
            },
            "/ms/vla/vla543210.ms": {
                "type": "google",
                "file_id": "123456",
                "hash": "123456",
                "filename": "vla012345.ms",
            },
        },
        schema=REGISTRY_SCHEMA)


def test_user_registry(registry):
    user_key = '/test/ms/2020-06-04/google/test_user_registry_ms.tar.gz'

    assert registry[user_key] == USER_REGISTRY_CONTENT[user_key]
